/**
Routines for graphs (arrows and boxes).
*/

import style;

usepackage("minibox");

string make_empty_label(int) { return ""; }
pen return_defaultpen(int) { return defaultpen; }
pen return_white_pen(int) { return white; }

pair centre(path p) {
    return .5*max(p) + .5*min(p);
}

pair extend_in_direction(pair offset, real distance) {
    return offset + unit(offset) * distance;
}

/**
Return a transform that rotates an object so that horizontal lines go in
"direction".
*/
transform rotate_to_direction(pair direction) {
    return (0, 0, direction.x, -direction.y, direction.y, direction.x);
}

/**
Take a left turn from start->end.
*/
pair orthogonal_direction(pair direction) {
    return (-direction.y, direction.x);
}

pair orthogonal_direction(path start, path end) {
    return orthogonal_direction(unit(centre(end) - centre(start)));
}

pair point_on_outside(path object, pair direction) {
    pair object_size = max(object) - min(object);
    real max_length = object_size.x + object_size.y;

    path line =
        centre(object) -- (centre(object) + unit(direction) * max_length);
    
    return intersectionpoint(object, line);
}

path combine(path o1, path o2) {
    return o1--o2;
}

path combine(path[] objects) {
    path result;
    for (int i=0; i < objects.length; ++i) {
        result = combine(result, objects[i]);
    }
    return result;
}


path box_around(path outline) {
    pair bottom_left = min(outline);
    pair top_right = max(outline);
    path outline = bottom_left -- (bottom_left.x, top_right.y)
        -- top_right -- (top_right.x, bottom_left.y) -- cycle;
    return outline;
}

/**
Remove the parts of line that go through "start" and "end".
*/
path cut_line_from_to(path line, path start, path end) {
    path first_to_middle = firstcut(line, start).after;
    return lastcut(first_to_middle, end).before;
}

path line_from_to(path start, path end, bool axis_aligned=false,
    real offset=0., real parallel=0.)
{
    pair start_centre = centre(start);
    pair end_centre = centre(end);
    pair direction = unit(end_centre - start_centre);
    pair centre_of_gravity = .5 * (start_centre + end_centre);
    if (axis_aligned) {
        if (direction.x < -sqrt(.5) || direction.x > sqrt(.5)) {
            // horizontal.
            start_centre = (start_centre.x, centre_of_gravity.y);
            end_centre = (end_centre.x, centre_of_gravity.y);
        } else {
            start_centre = (centre_of_gravity.x, start_centre.y);
            end_centre = (centre_of_gravity.x, end_centre.y);
        }
    }
    pair orthogonal = orthogonal_direction(start, end);
    pair middle = .5 * (start_centre + end_centre) + offset * orthogonal;
    path complete_line = shift(parallel * orthogonal)
        * (start_centre .. middle .. end_centre);
    return cut_line_from_to(complete_line, start, end);
}

path manhattan_line_from_to(pair start, pair end, pair start_direction=N,
    real corner_from_start=-1, real corner_from_end=-1)
{
    path middle_segment;
    start_direction = unit(start_direction);


    real distance_in_direction = dot(end - start, start_direction);
    if (corner_from_start == -1) {
        if (corner_from_end == -1) {
            // Pick middle.
            corner_from_end = distance_in_direction / 2;
        }
        corner_from_start = distance_in_direction - corner_from_end;
    } else {
        corner_from_end = distance_in_direction - corner_from_start;
    }

    path middle_segment = (start + start_direction * corner_from_start)
        -- (end - start_direction * corner_from_end);

    return start--middle_segment--end;
}

path manhattan_line_from_to(path start, path end, pair start_direction=N,
    real corner_from_start=-1, real corner_from_end=-1)
{
    return manhattan_line_from_to(
        point_on_outside(start, start_direction),
        point_on_outside(end, -start_direction),
        start_direction=start_direction, corner_from_start=corner_from_start,
        corner_from_end=corner_from_end);
}

/*** FloatingObject. ***/

typedef path DrawTransformed(picture, transform);
path draw_nothing(picture pic, transform) { return nullpath; }

/// Object that can be placed in a different position.
struct FloatingObject {
    path outline;
    DrawTransformed draw_transformed;

    /// Initialise empty.
    static FloatingObject FloatingObject() {
        FloatingObject o = new FloatingObject;
        o.outline = nullpath;
        o.draw_transformed = draw_nothing;
        return o;
    }

    /// Initialise with the given values.
    static FloatingObject FloatingObject(path outline, DrawTransformed draw_transformed) {
        FloatingObject o = new FloatingObject;
        o.outline = outline;
        o.draw_transformed = draw_transformed;
        return o;
    }

    /// Initialise from simple outline.
    static FloatingObject from_outline(path outline, pen p=defaultpen, pen fillpen=invisible)
    {
        path draw_transformed(picture pic, transform t) {
            path shape = t * outline;
            fill(shape, fillpen);
            draw(shape, p);
            return shape;
        }
        FloatingObject o = new FloatingObject;
        o.outline = outline;
        o.draw_transformed = draw_transformed;
        return o;
    }

    FloatingObject apply_transform(transform t) {
        path outline = t * this.outline;
        path draw_transformed(picture pic, transform t2) {
            return this.draw_transformed(pic, t2 * t);
        }
        return FloatingObject(outline, draw_transformed);
    }

    FloatingObject shift_by(pair location) {
        return apply_transform(shift(location));
    }
};

from FloatingObject unravel FloatingObject;

FloatingObject operator*(transform t, FloatingObject o) {
    return o.apply_transform(t);
}

FloatingObject combine(
    pair position1=(0,0), FloatingObject o1,
    pair position2=(0,0), FloatingObject o2)
{
    path outline = (shift(position1) * o1.outline)
        -- (shift(position2) * o2.outline);
    path draw_transformed(picture pic, transform t) {
        o1.draw_transformed(pic, t * shift(position1));
        o2.draw_transformed(pic, t * shift(position2));
        return t * outline;
    }
    return FloatingObject(outline, draw_transformed);
}

FloatingObject combine(FloatingObject[] objects) {
    FloatingObject result = FloatingObject();
    for (int i=0; i < objects.length; ++i) {
        result = combine(result, objects[i]);
    }
    return result;
}

FloatingObject make_invisible(FloatingObject o) {
    path draw_invisible(picture, transform t) {
        picture void_;
        return o.draw_transformed(void_, t);
    }

    return FloatingObject(o.outline, draw_invisible);
}

/// Replace the outline of "o" by its bounding box.
FloatingObject box_outline(FloatingObject o) {
    return FloatingObject(box_around(o.outline), o.draw_transformed);
}

/// Compute how "o" should be moved to be "gap" away from "reference".
pair location_next_to(path reference, pair gap, FloatingObject o,
    bool reference_bounding_box=false, bool object_bounding_box=false)
{
    if (reference_bounding_box)
        reference = box_around(reference);
    pair reference_point = point_on_outside(reference, unit(gap));

    path object_outline = o.outline;
    if (object_bounding_box)
        object_outline = box_around(object_outline);
    pair floating_point = point_on_outside(object_outline, -unit(gap));
    return reference_point + gap - floating_point;
}

/// Move "o" so that it is "gap" away from "reference".
FloatingObject next_to(path reference, pair gap, FloatingObject o,
    bool reference_bounding_box=false, bool object_bounding_box=false)
{
    return shift(location_next_to(reference, gap, o,
        reference_bounding_box=reference_bounding_box,
        object_bounding_box=object_bounding_box)) * o;
}

/// Move "os" so that they are "gap" away from "reference".
FloatingObject[] next_to(path reference,
    pair gap, FloatingObject[] os,
    bool reference_bounding_box=false, bool object_bounding_box=false)
{
    pair location = location_next_to(reference, gap, combine(os),
        reference_bounding_box=reference_bounding_box,
        object_bounding_box=object_bounding_box);
    FloatingObject[] result;
    for (int i = 0; i < os.length; ++ i)
        result.push(shift(location) * os[i]);
    return result;
}

/// Move "o" so that it is "gap" away from "reference".
FloatingObject next_to(FloatingObject reference, pair gap, FloatingObject o,
    bool reference_bounding_box=false, bool object_bounding_box=false)
{
    return next_to(reference.outline, gap, o,
        reference_bounding_box=reference_bounding_box,
        object_bounding_box=object_bounding_box);
}

/**
Move object "o" so that a point on its outside is in the same place
as the equivalent point on the reference, measured in "direction".
E.g. to line up object o2 with o2 at the top, say "align(o1, N, o2)".
*/
FloatingObject align(path reference, pair reference_direction,
    pair measure_direction=reference_direction, FloatingObject o,
    bool reference_bounding_box=false)
{
    if (reference_bounding_box)
        reference = box_around(reference);
    pair reference_point = point_on_outside(reference, unit(reference_direction));
    pair floating_point = point_on_outside(o.outline, unit(reference_direction));
    real distance = dot(unit(measure_direction), reference_point - floating_point);
    return shift(distance * unit(measure_direction)) * o;
}

/// Draw a FloatingObject at a specified location and return its outline.
path draw_at(picture pic=currentpicture, pair position, FloatingObject o) {
    return o.draw_transformed(pic, shift(position));
}

/**
Draw multiple FloatinObject's at a specified location and return their outlines.
*/
path[] draw_at(picture pic=currentpicture, pair position, FloatingObject[] os) {
    path[] result;
    for (int i = 0; i < os.length; ++ i)
        result.push(os[i].draw_transformed(pic, shift(position)));
    return result;
}

/// Draw a floating object at a specified distance to a path.
path draw_next_to(picture pic=currentpicture, path reference, pair gap,
    FloatingObject o,
    bool reference_bounding_box=false, bool object_bounding_box=false)
{
    return draw_at(pic, (0,0), next_to(reference, gap, o,
        reference_bounding_box=reference_bounding_box,
        object_bounding_box=object_bounding_box));
}

/**
Draw multiple floating objects at a specified distance to a path.
If use_bounding_box=true, then the bounding box is used to figure out the
distance.
This is often useful for odd numbers of objects.
*/
path[] draw_next_to(picture pic=currentpicture, path reference,
    pair gap, FloatingObject[] os,
    bool reference_bounding_box=false, bool object_bounding_box=false)
{
    pair location = location_next_to(reference, gap, combine(os),
        reference_bounding_box=reference_bounding_box,
        object_bounding_box=object_bounding_box);
    return draw_at(pic, location, os);
}

/*** Shapes. ***/

path circle_shape(pair centre, real radius) {
    return shift(centre) * scale(radius) * unitcircle;
}

path rectangle(pair centre, pair size) {
    path unit_square = (0,0) -- (0,1) -- (1,1) -- (1,0) -- cycle;
    return shift(centre) *  scale(size.x, size.y) * shift((-.5, -.5))
        * unit_square;
}

path rounded_rectangle_shape(pair centre, pair size, real radius) {
    if (size.x/3 < radius) {
        radius = size.x/3;
    }
    if (size.y/3 < radius) {
        radius = size.y/3;
    }

    // Corners should not be exact fragments of circles, but smoothly
    // transition.
    real corner_start = 1.5;
    real corner_waypoint_offset = .05 * radius;

    path corner_bottom_left = (corner_start * radius, 0){left}..
        (radius, corner_waypoint_offset)..
        (corner_waypoint_offset, radius)..
        (0, corner_start * radius){up};
    path corner_top_left = shift((0, size.y)) * scale(1, -1)
        * corner_bottom_left;

    // Assemble left side.
    path left_side = corner_bottom_left--reverse(corner_top_left);
    // Mirror left size to make right side.
    path right_side = shift((size.x, 0)) * scale(-1, 1) * left_side;

    // Connect left and right side.
    path result = left_side -- reverse(right_side) -- cycle;

    return shift(centre) * shift(-.5*size) * result;
}

/// Convert a desired gap between objects to a translation for o2 to obtain the
/// required gap.
pair gap_to_step(FloatingObject o1, FloatingObject o2=o1, pair gap) {
    pair reference_point = point_on_outside(o1.outline, unit(gap));
    pair next_point = point_on_outside(o2.outline, -unit(gap));
    return reference_point + gap - next_point;
}

/** Drawing routines. **/

path actually_draw_arrow(picture pic=currentpicture, path line,
    pen p=arrow_pen, real arrow_size=default_arrow_size,
    string label_text="", pair label_align=(0,0), real label_position=0.5)
{
    draw(pic, line, p, EndArrow(size=arrow_size));
    if (label_text != "") {
        if (label_align == (0,0))
            label_align = orthogonal_direction(dir(line, label_position));
        label(pic, Label(label_text, relpoint(line, label_position)),
            align=label_align, p);
    }
    return line;
}

path draw_arrow(picture pic=currentpicture,
    path start, path end, bool axis_aligned=false,
    real offset = 0., real parallel = 0.,
    pen p = arrow_pen, real arrow_size = default_arrow_size,
    string label_text = "", pair label_align = (0,0),
    real label_position = 0.5)
{
    path line = line_from_to(start, end, axis_aligned=axis_aligned,
        offset=offset, parallel=parallel);
    return actually_draw_arrow(pic=pic, line=line, p=p, arrow_size=arrow_size,
        label_text=label_text, label_align=label_align,
        label_position=label_position);
}

path draw_manhattan_arrow(picture pic=currentpicture, path start, path end,
    pair start_direction=N,
    real corner_from_start=-1, real corner_from_end=-1,
    pen p = arrow_pen, real arrow_size = default_arrow_size,
    string label_text = "", pair label_align = (0,0),
    real label_position = 0.5)
{
    path line = manhattan_line_from_to(
        start=start,
        end=end,
        start_direction=start_direction,
        corner_from_start=corner_from_start,
        corner_from_end=corner_from_end);
    return actually_draw_arrow(pic=pic, line=line, p=p, arrow_size=arrow_size,
        label_text=label_text, label_align=label_align,
        label_position=label_position);
}

FloatingObject circle(real radius, pen p = defaultpen, pen fillpen=invisible,
    string label_text = "")
{
    path outline = circle_shape((0,0), radius);
    path draw_transformed(picture pic, transform t) {
        path shape = t * outline;
        fill(shape, fillpen);
        if (label_text != "")
            label(pic, t * Label(label_text), p);
        draw(shape, p);
        return shape;
    }
    return FloatingObject(outline, draw_transformed);
}

path draw_circle(
    picture pic=currentpicture,
    pair centre, real radius, pen p = defaultpen, string label_text = "")
{
    path result = circle_shape(centre, radius);
    draw(pic, result, p);
    if (label_text != "")
        label(pic, Label(label_text, centre), p);
    return result;
}

void draw_width_constrained_label(picture pic=currentpicture,
    transform t, real width, string label_text, pen p)
{
    if (label_text != "") {
        label(pic, t * Label(
            "\parbox{" + string(width) +"pt}{\centering{" + label_text + "}}"),
            p);
    }
}

path draw_rectangle(picture pic=currentpicture,
    pair centre, pair size,
    pen p = defaultpen, pen fillpen=invisible, string label_text = "")
{
    path result = rectangle(centre, size);
    fill(pic, result, fillpen);
    draw(pic, result, p);
    if (label_text != "")
        label(pic, Label(label_text, centre), p);
    return result;
}

path draw_text(picture pic=currentpicture,
    pair centre, pair size=unboxed_size,
    pen p = defaultpen, string label_text)
{
    label(pic, Label(label_text, centre), p);
    return rectangle(centre, size);
}

FloatingObject rounded_rectangle(
    pair size, real radius,
    pen p=defaultpen, pen fillpen=invisible,
    string label_text = "", pen label_pen=p)
{
    path floating_shape = rounded_rectangle_shape((0,0), size, radius);
    path draw_at(picture pic, transform t) {
        path shape = t * floating_shape;
        fill(pic, shape, fillpen);
        draw(pic, shape, p);
        draw_width_constrained_label(pic, t, size.x, label_text, label_pen);
        return shape;
    }

    return FloatingObject(floating_shape, draw_at);
}

FloatingObject just_text(
    pair size, real radius,
    pen p = defaultpen, string label_text = "")
{
    path floating_shape = rounded_rectangle_shape((0,0), size, radius);
    path draw_at(picture pic, transform t) {
        path shape = t * floating_shape;
        draw_width_constrained_label(pic, t, size.x, label_text, p);
        return shape;
    }

    return FloatingObject(floating_shape, draw_at);
}

// TODO Backward compatibility.
path draw_rounded_rectangle(
    picture pic=currentpicture,
    pair centre, pair size, real radius,
    pen p = defaultpen, pen fillpen=invisible, string label_text = "")
{
    return draw_at(pic, centre, rounded_rectangle(
        size=size, radius=radius, p=p, fillpen=fillpen, label_text=label_text));
}

/// Decide on an alignment for text, based on the direction.
/// -1: left; 0: center; 1: right.
int text_alignment_from_direction(pair direction) {
    if (unit(direction).x > sqrt(.5)) {
        // Label mostly to the right.
        return -1;
    } else if (unit(direction).x < -sqrt(.5)) {
        // Label mostly to the left.
        return +1;
    }
    return 0;
}

/**
Add a label at a point, with some features for layout.
The alignment is chosen depending on the direction.
If the width is given, a parbox is put around the text.
If not, the width is determined automatically.
Either way, newlines can be used.
*/
void label_point(picture pic=currentpicture,
    pair point, string label, pair direction=N, pen p=defaultpen,
    real width=0., real rotation=0.)
{
    string boxed_label;
    int alignment = text_alignment_from_direction(
        rotate(-rotation) * direction);

    if (width == 0.) {
        string alignment_text;
        if (alignment == -1) {
            alignment_text = "l";
        } else if (alignment == +1) {
            alignment_text = "r";
        } else {
            alignment_text = "c";
        }

        boxed_label = "\minibox[" + alignment_text + "]{" + label + "}";
    } else {
        string alignment_text;
        if (alignment == -1) {
            alignment_text = "\raggedright";
        } else if (alignment == +1) {
            alignment_text = "\raggedleft";
        } else {
            alignment_text = "\centering";
        }

        boxed_label = "\parbox{" + string(width) + "pt}{" + alignment_text + " "
            + label + "}";
    }
    label(pic,
        rotate(rotation) * Label(boxed_label, point), align=direction, p);
}

void label_outside(picture pic=currentpicture,
    path object, string label, pair direction=N, pen p=defaultpen,
    real width=0., real rotation=0.)
{
    pair point = point_on_outside(object, direction);
    label_point(pic, point, label, direction=direction, p=p,
        width=width, rotation=rotation);
}

void label_inside(picture pic=currentpicture,
    path object, string label, pair direction=N, pen p=defaultpen,
    real width=0., real rotation=0.)
{
    pair point = point_on_outside(object, direction);
    label_point(pic, point, label, direction=-direction, p=p,
        width=width, rotation=rotation);
}

/*** Flow diagrams. ***/

FloatingObject process_box(string label_text="",
    int num_lines=1, real width_factor=1, pen p=defaultpen)
{
    pair size = process_size_for_lines(
        num_lines=num_lines, width_factor=width_factor);
    return rounded_rectangle(size, process_radius, p=p, fillpen=white,
        label_text=label_text);
}

FloatingObject text_box(string label_text="",
    int num_lines=1, real width_factor=1, pen p=defaultpen)
{
    // TODO Height is wrong.
    pair size = process_size_for_lines(
        num_lines=num_lines, width_factor=width_factor);
    return just_text(size, process_radius, p=p, label_text=label_text);
}

/*** Plates. ***/

FloatingObject plate_around(FloatingObject o,
    real margin=plate_margin,
    real left_margin=-1, real right_margin=-1,
    real bottom_margin=-1, real top_margin=-1,
    real corner_radius=margin,
    pen p=shade_box_label_pen, pen fillpen=make_shade_box_pen(p))
{
    if (left_margin < 0)
        left_margin = margin;
    if (right_margin < 0)
        right_margin = margin;
    if (bottom_margin < 0)
        bottom_margin = margin;
    if (top_margin < 0)
        top_margin = margin;
    
    pair bottom_left = min(o.outline) - (left_margin, bottom_margin);
    pair top_right = max(o.outline) + (right_margin, top_margin);

    pair size = top_right - bottom_left;
    pair position = .5 * (bottom_left + top_right);

    return rounded_rectangle(size, radius = corner_radius,
        p=p, fillpen=fillpen).shift_by(position);
}

/// TODO temporary for backward compatibility.

path draw_plate_around(picture pic=currentpicture,
    path element,
    real margin=plate_margin,
    real left_margin=-1, real right_margin=-1,
    real bottom_margin=-1, real top_margin=-1,
    pen p=shade_box_label_pen, pen fillpen=make_shade_box_pen(p))
{
    return draw_at(pic, (0,0),
        plate_around(
            FloatingObject(element, draw_nothing), // !!! Hack.
            margin=margin,
            left_margin=left_margin, right_margin=right_margin,
            bottom_margin=bottom_margin, top_margin=top_margin,
            p=p, fillpen=fillpen
        ));
}

path draw_plate_around(picture pic=currentpicture,
    path[] elements,
    real margin=plate_margin,
    real left_margin=-1, real right_margin=-1,
    real bottom_margin=-1, real top_margin=-1,
    pen p=shade_box_label_pen, pen fillpen=make_shade_box_pen(p))
{
    FloatingObject[] objects;
    for (int i=0; i < elements.length; ++ i)
        objects.push(FloatingObject(elements[i], draw_nothing)); // !!! Hack.
    return draw_at(pic, (0,0),
        plate_around(
            combine(objects),
            margin=margin,
            left_margin=left_margin, right_margin=right_margin,
            bottom_margin=bottom_margin, top_margin=top_margin,
            p=p, fillpen=fillpen
        ));
}

FloatingObject with_plate_around(FloatingObject o,
    real margin=plate_margin,
    real left_margin=-1, real right_margin=-1,
    real bottom_margin=-1, real top_margin=-1,
    real corner_radius=margin,
    pen p=shade_box_label_pen, pen fillpen=make_shade_box_pen(p))
{
    FloatingObject plate = plate_around(o, margin=margin,
        left_margin=left_margin,
        right_margin=right_margin,
        bottom_margin=bottom_margin,
        top_margin=top_margin,
        corner_radius=corner_radius,
        p=p,
        fillpen=fillpen);

    path draw_transformed(picture pic, transform t) {
        path result = plate.draw_transformed(pic, t);
        o.draw_transformed(pic, t);
        return result;
    }
    return FloatingObject(plate.outline, draw_transformed);
}

/** Finite-state specific. **/

path finite_state(pair centre, real radius = finite_state_radius)
{ return circle(centre, radius); }

path draw_finite_state(
    picture pic=currentpicture,
    pair centre,
    real radius = finite_state_radius,
    pen p = finite_state_pen,
    string label_text = "")
{ return draw_circle(pic, centre, radius, p, label_text); }

path draw_finite_end_state(
    picture pic=currentpicture,
    pair centre,
    real inner_radius = finite_state_inner_radius,
    real outer_radius = finite_state_outer_radius,
    pen p = finite_state_pen,
    string label_text = "")
{
    // Text in inner circle.
    draw_circle(pic, centre, inner_radius, p, label_text);
    // Return outer circle.
    return draw_circle(pic, centre, outer_radius, p, "");
}

/** Neural-net specific. **/

/// Neural network vector (with the right corner radius).
FloatingObject vector(pair size=vector_size,
    pen p=defaultpen, pen fillpen=white, string label_text="")
{
    return rounded_rectangle(size, radius=vector_corner_radius,
        p=p, fillpen=fillpen, label_text=label_text);
}

FloatingObject[] vector_sequence(
    int number, pair size=vector_size, pair gap=time_gap,
    bool around_origin=false,
    pen make_pen(int) = return_defaultpen,
    pen make_fillpen(int) = return_white_pen,
    string make_label(int) = make_empty_label)
{
    FloatingObject[] result;
    pair step = gap_to_step(vector(size), gap);
    pair current_position = (0,0);
    if (around_origin)
        current_position = -(number-1)/2 * step;
    
    for (int i = 0; i < number; ++ i) {
        result.push(vector(
            size,
            make_pen(i), fillpen=make_fillpen(i), make_label(i))
            .shift_by(current_position));
        current_position += step;
    }
    return result;
}

FloatingObject vector_sequence_on_plate(int number,
    string make_vector_label(int),
    pen label_pen, pen fillpen)
{
    FloatingObject vectors = combine(vector_sequence(
        number=number, make_vector_label));

    return with_plate_around(
        vectors, p=label_pen, fillpen=fillpen);
}

FloatingObject circular_operator(real size=atomic_operation_radius,
    pen p, pen fillpen,
    void draw_inside(picture,
        transform transform_size_and_location, transform transform_location))
{
    transform transform_size = scale(size);
    path circle_shape = transform_size * unitcircle;

    path draw_transformed(picture pic, transform t) {
        path circle = t * circle_shape;
        fill(circle, fillpen);
        draw_inside(pic, t * transform_size, t);
        draw(circle, p);
        return circle;
    }

    return FloatingObject(circle_shape, draw_transformed);
}

FloatingObject cross_box(real size=atomic_operation_radius, real blade_rotation,
    pen p, pen fillpen)
{
    transform rotation = rotate(blade_rotation);
    void draw_inside(picture, transform t, transform) {
        draw (t * rotation * ((-1, 0)--(+1, 0)), p);
        draw (t * rotation * ((0, -1)--(0, +1)), p);
    }
    return circular_operator(size=size, p=p, fillpen=fillpen, draw_inside);
}

FloatingObject times_box(real size=atomic_operation_radius,
    pen p=defaultpen, pen fillpen=white)
{
    return cross_box(size=size, blade_rotation=45, p=p, fillpen=fillpen);
}

FloatingObject plus_box(real size=atomic_operation_radius,
    pen p=defaultpen, pen fillpen=white)
{
    return cross_box(size=size, blade_rotation=0, p=p, fillpen=fillpen);
}

FloatingObject sigmoid_box(real size=atomic_operation_radius,
    pen p=defaultpen, pen fillpen=white)
{
    void draw_inside(picture pic, transform t, transform) {
        real symmetric_sigmoid(real x) {
            real exp_x = exp(x);
            return (exp_x - 1/exp_x) / (exp_x + 1/exp_x);
        }

        // 2^2 = extent^2 + 1^2
        real extent = sqrt(3);
        
        guide g;
        for (real x= -extent; x <= extent; x += extent/9.999) {
            g = g..(x, symmetric_sigmoid(2*x));
        }

        draw(pic, t * scale(.5) * ((-extent, -1) -- (+extent, -1)),
            linewidth(0.5 * linewidth(p)) + dotted);

        // Draw horizontal axis at 0.
        draw(pic, t * scale(.5) * g, p);
    }
    return circular_operator(size=size, p=p, fillpen=fillpen, draw_inside);
}

FloatingObject clipped_circular_operator(real size=atomic_operation_radius,
    pen p, pen fillpen,
    void draw_inside(picture, path,
        transform transform_size_and_location, transform transform_location))
{
    transform transform_size = scale(size);
    path circle_shape = transform_size * unitcircle;

    path draw_transformed(picture pic, transform t) {
        path circle = t * circle_shape;
        fill(circle, fillpen);
        draw_inside(pic, circle, t * transform_size, t);
        draw(circle, p);
        return circle;
    }

    return FloatingObject(circle_shape, draw_transformed);
}


FloatingObject soft_relu_box(real size=atomic_operation_radius,
    pen p=defaultpen, pen fillpen=white)
{
    void draw_inside(picture pic, path border, transform t, transform) {
        real zoom = 3;
        real soft_relu(real x) {
            return log(1 + exp(zoom * x))/zoom;
        }

        picture inside;

        real extent = 5;
        transform s = shift((0, -.5)) * scale(2/extent);
        
        guide g;
        for (real x= -extent; x <= extent; x += .01) {
            g = g..(x, soft_relu(x));
        }

        draw(inside, t * s * ((-extent, 0) -- (+extent, 0)),
            linewidth(0.5 * linewidth(p)) + dotted);

        draw(inside, t * s * g, p);

        clip(inside, border);

        add(pic, inside);
    }
    return clipped_circular_operator(size=size, p=p, fillpen=fillpen, draw_inside);
}

FloatingObject soft_relu_box_old(real size=atomic_operation_radius,
    pen p=defaultpen, pen fillpen=white)
{
    void draw_inside(picture pic, transform t, transform) {
        real soft_relu(real x) {
            return log(1 + exp(x));
        }

        // 2^2 = extent^2 + 1^2
        real extent = sqrt(3);
        
        guide g;
        for (real x= -extent; x <= extent; x += extent/9.999) {
            g = g..(x, soft_relu(2*x));
        }

        draw(pic, t * scale(.5) * ((-extent, -1) -- (+extent, -1)),
            linewidth(0.5 * linewidth(p)) + dotted);

        // Draw horizontal axis at 0.
        draw(pic, t * scale(.5) * g, p);
    }
    return circular_operator(size=size, p=p, fillpen=fillpen, draw_inside);
}

FloatingObject matmul_box(real size=atomic_operation_radius,
    pen p=defaultpen, pen fillpen=white)
{
    void draw_inside(picture pic, transform, transform t) {
        label(pic, t * Label("$\cdot$"), p);
    }
    return circular_operator(size=size, p=p, fillpen=fillpen, draw_inside);
}
